/*
 * Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "uioutils.h"
#include "mpmloc.h"

#define MAX_FILE_NAME_LENGTH	128
#define MAX_NAME_LENGTH		64
#define MAX_LINE_LENGTH         16

int endian_check() {
    int check = 1;
    char ret = *(char *) &check;
    return ret;
}

int byte_swap(int x) {
	return ( ((x>>0) & 0xff) << 24)
		 | ( ((x>>8) & 0xff) << 16)
		 | ( ((x>>16) & 0xff) << 8)
		 | ( ((x>>24) & 0xff) << 0);
}

int uioutil_get_string (char *filename, char *str, int str_len)
{
	FILE *fpr = 0;
	int ret_val = -1;
	char *strptr;

	if (!filename || !str || !str_len) {
		goto close_n_exit;
	}

	fpr = fopen(filename, "r");
	if (!fpr) {
		error_msg("Error opening file %s (%s)",
		       filename, strerror(errno));
		goto close_n_exit;
	}
	if (!fgets(str, str_len, fpr)) {
		error_msg("Error reading file %s (%s)",
		       filename, strerror(errno));
		goto close_n_exit;
	}

	/* Remove new line from string */
	if (!(strptr = strchr(str, '\n'))) {
		error_msg("Too long for buffer %s", str);
		goto close_n_exit;
	}
	*strptr = '\0';

	ret_val = 0;

close_n_exit:
	if (fpr) fclose(fpr);
	return ret_val;
}

int uioutil_get_int (char *filename, int *val)
{
	FILE *fpr = 0;
	int ret_val = -1;
	char line[MAX_LINE_LENGTH];

	if (!filename || !val)
		return -1;

	if (uioutil_get_string(filename, line, MAX_LINE_LENGTH) < 0)
		return -1;

	*val = strtoul(line, NULL, 0);
	return 0;
}
char  *remove_prefix_from_device_name(char *name, int max_length)
{ 
  int index;

  index=0;
  
  /* strip anything before . in the name */
  while(name[index] != '\0')
  {
    if(name[index] == '.')
    { 
      index++;
      return(&name[index]);
    }
    if(index > max_length)
      break;
    index++;
  }
  return(name);
}
int uioutil_get_device (char *uio_name, char *class_name, int class_name_length)
{
	struct dirent *entry = 0;
	char filename[MAX_FILE_NAME_LENGTH];
	char name[MAX_NAME_LENGTH];
	int ret_val = -1;
	DIR *dir = 0;
        char *name_ptr;

	if (!uio_name || !class_name || !strlen(uio_name)) {
		goto close_n_exit;
	}

	dir = opendir("/sys/class/uio");
	if (!dir) {
		perror("readdir /sys/class/uio");
		goto close_n_exit;
	}

	while ((entry = readdir(dir)) != NULL) {
		if (strstr (entry->d_name, "uio") == NULL) {
			continue;
		}

		snprintf(filename, MAX_FILE_NAME_LENGTH, "/sys/class/uio/%s/name", entry->d_name);

		if (uioutil_get_string(filename, name, MAX_NAME_LENGTH) < 0)
			goto close_n_exit;
                name_ptr = remove_prefix_from_device_name(name, MAX_NAME_LENGTH);
		if (!strncmp (uio_name, name_ptr, strlen(uio_name))) {
			ret_val = 0;
			strncpy (class_name, entry->d_name, class_name_length);
			goto close_n_exit;
		}
	}

close_n_exit:
	if (dir) closedir(dir);

	return ret_val;
}

int uioutil_mapinfo_from_addr (uioutil_map_info_t *p_info, char *class_name, int address, int size)
{
	struct dirent *entry = 0;
	char filename[MAX_FILE_NAME_LENGTH];
	char dirname[MAX_FILE_NAME_LENGTH];
	uint32_t base_size;
	int ret_val = -1;
	uint32_t base;
	DIR *dir = 0;
	int index;

	if (!p_info || !class_name) {
		goto close_n_exit;
	}

	snprintf(dirname, MAX_FILE_NAME_LENGTH, "/sys/class/uio/%s/maps", class_name);
	dir = opendir(dirname);
	if (!dir) {
		error_msg("Error opening directory %s (%s)",
		       dirname, strerror(errno));
		goto close_n_exit;
	}

	while ((entry = readdir(dir)) != NULL) {

		if (!strstr (entry->d_name, "map")) {
			continue;
		}

		snprintf(filename, MAX_FILE_NAME_LENGTH, "%s/%s/addr", dirname, entry->d_name);
		if (uioutil_get_int(filename, &base) < 0) {
			goto close_n_exit;
		}

		snprintf(filename, MAX_FILE_NAME_LENGTH, "%s/%s/size", dirname, entry->d_name);
		if (uioutil_get_int(filename, &base_size) < 0) {
			goto close_n_exit;
		}

		if (base <= address && base + base_size >= address + size) {

			snprintf(filename, MAX_FILE_NAME_LENGTH, "%s/%s/offset", dirname, entry->d_name);
			if (uioutil_get_int(filename, &p_info->offset) < 0) {
				goto close_n_exit;
			}

			if (sscanf(entry->d_name, "map%d", &index) != 1) {
				error_msg("Error in scanning index from %s (%s)",
				       entry->d_name, strerror(errno));
				goto close_n_exit;
			}
			p_info->base = base;
			p_info->size = base_size;
			p_info->index = index;

			ret_val = 0;
			goto close_n_exit;
		}

	}

close_n_exit:
	if (dir) closedir(dir);

	return ret_val;
}

int uioutil_offset_from_proc (char *dev_class, int index)
{
	char filename[MAX_FILE_NAME_LENGTH];
	char dsp_name[MAX_FILE_NAME_LENGTH];
	int i;
	int ret;
	FILE *fp;
	char *name_ptr;

	snprintf(filename, MAX_FILE_NAME_LENGTH, "/sys/class/uio/%s/name", dev_class);

	if (uioutil_get_string(filename, dsp_name, MAX_FILE_NAME_LENGTH) < 0) {
		return(-1);
	}
	name_ptr = remove_prefix_from_device_name(dsp_name, MAX_FILE_NAME_LENGTH);

	snprintf(filename, MAX_FILE_NAME_LENGTH, "/proc/device-tree/soc/%s/reg", name_ptr);

	fp = fopen(filename, "rb");
	if (!fp) return -1;
	for (i=0; i<(index*2 + 1); i++) {
		fread(&ret, sizeof(int), 1, fp);
	}
	fclose(fp);
	if (endian_check()) ret = byte_swap(ret);
	return ret & (getpagesize() - 1);
}

int uioutil_mapinfo_from_name (uioutil_map_info_t *p_info, char *dev_class, char *name)
{
	struct dirent *entry = 0;
	char filename[MAX_FILE_NAME_LENGTH];
	char dirname[MAX_FILE_NAME_LENGTH];
	char name_str[MAX_NAME_LENGTH];
	int ret_val = -1;
	DIR *dir = 0;
	int index;
	int offset;

	if (!p_info || !dev_class || !name) {
		goto close_n_exit;
	}

	snprintf(dirname, MAX_FILE_NAME_LENGTH, "/sys/class/uio/%s/maps", dev_class);
	dir = opendir(dirname);
	if (!dir) {
		error_msg("Error opening directory %s (%s)",
		       dirname, strerror(errno));
		goto close_n_exit;
	}

	while ((entry = readdir(dir)) != NULL) {

		if (!strstr (entry->d_name, "map")) {
			continue;
		}

		snprintf(filename, MAX_FILE_NAME_LENGTH, "%s/%s/name", dirname, entry->d_name);
		if (uioutil_get_string(filename, name_str, MAX_NAME_LENGTH) < 0) {
			goto close_n_exit;
		}

		if (strncmp (name_str, name, strlen(name))) {
			continue;
		}

		snprintf(filename, MAX_FILE_NAME_LENGTH, "%s/%s/addr", dirname, entry->d_name);
		if (uioutil_get_int(filename, &p_info->base) < 0) {
			goto close_n_exit;
		}

		snprintf(filename, MAX_FILE_NAME_LENGTH, "%s/%s/size", dirname, entry->d_name);
		if (uioutil_get_int(filename, &p_info->size) < 0) {
			goto close_n_exit;
		}

		snprintf(filename, MAX_FILE_NAME_LENGTH, "%s/%s/offset", dirname, entry->d_name);
		if (uioutil_get_int(filename, &p_info->offset) < 0) {
			goto close_n_exit;
		}

		if (sscanf(entry->d_name, "map%d", &index) != 1) {
			error_msg("Error in scanning index from %s (%s)",
			       entry->d_name, strerror(errno));
			goto close_n_exit;
		}

		p_info->index = index;

		offset = uioutil_offset_from_proc(dev_class, index);
		if (p_info->offset != offset)
			p_info->offset = offset;

		ret_val = 0;
		break;

	}

close_n_exit:
	if (dir) closedir(dir);

	return ret_val;

}
#if 0
int uioutil_write_reg (char *class_name, char *uio_name, int val, uioutil_oper_t op)
{
	char dev_name[MAX_FILE_NAME_LENGTH];
	uioutil_map_info_t map_info;
	int ret_val = -1;
	char *map = 0;
	char *regp;
	int fd = 0;

	if (uioutil_mapinfo_from_name (&map_info, class_name, uio_name) < 0) {

		error_msg("Failed to get mapinfo for name %s class %s",
		       uio_name, class_name);
		goto close_n_exit;
	}

	snprintf(dev_name, MAX_FILE_NAME_LENGTH, "/dev/%s", class_name);
	fd = open(dev_name, (O_RDWR | O_SYNC));
	if (fd < 0) {
		error_msg("Failed to open %s (%s)",
		       dev_name, strerror(errno));
		goto close_n_exit;
	}

	map = mmap(NULL, map_info.size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, map_info.index * getpagesize());
	if (!map) {
		error_msg("UIO map failed for %s (%s)",
		       dev_name, strerror(errno));
		goto close_n_exit;
	}
        debug_msg(" map info offset %x", map_info.offset);
	regp = map + map_info.offset;
        debug_msg(" Reg pointer for %x: %p", map_info.base, regp);

	switch (op) {
	case assign_op:
		*(uint32_t *)regp = val;;
		break;
	case and_op:
		*(uint32_t *)regp &= val;
		break;
	case or_op:
		*(uint32_t *)regp |= val;
		break;
	default:
		ret_val = -1;
		break;
	}

	ret_val = 0;
close_n_exit:
	if (fd > 0) close (fd);
	if (map > 0) munmap (map, map_info.size);

	return ret_val;
}
int uioutil_read_reg (char *class_name, char *uio_name, uint32_t *val)
{
	char dev_name[MAX_FILE_NAME_LENGTH];
	uioutil_map_info_t map_info;
	int ret_val = -1;
	char *map = 0;
	char *regp;
	int fd = 0;

	if (uioutil_mapinfo_from_name (&map_info, class_name, uio_name) < 0) {

		error_msg("Failed to get mapinfo for name %s class %s",
		       uio_name, class_name);
		goto close_n_exit;
	}

	snprintf(dev_name, MAX_FILE_NAME_LENGTH, "/dev/%s", class_name);
	fd = open(dev_name, (O_RDWR | O_SYNC));
	if (fd < 0) {
		error_msg("Failed to open %s (%s)",
		       dev_name, strerror(errno));
		goto close_n_exit;
	}

	map = mmap(NULL, map_info.size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, map_info.index * getpagesize());
	if (!map) {
		error_msg("UIO map failed for %s (%s)",
		       dev_name, strerror(errno));
		goto close_n_exit;
	}

	regp = map + map_info.offset;
     
        *val = *(uint32_t *)regp;

 	ret_val = 0;
close_n_exit:
	if (fd > 0) close (fd);
	if (map > 0) munmap (map, map_info.size);

	return ret_val;
}
#else
int uioutil_write_reg (char *class_name, char *uio_name, int fd, int val, uioutil_oper_t op)
{
	char dev_name[MAX_FILE_NAME_LENGTH];
	uioutil_map_info_t map_info;
	int ret_val = -1;
	char *map = 0;
	char *regp;

	if(fd == 0)
		return(ret_val);

	if (uioutil_mapinfo_from_name (&map_info, class_name, uio_name) < 0) {

		error_msg("Failed to get mapinfo for name %s class %s",
		       uio_name, class_name);
		goto close_n_exit;
	}

	map = mmap(NULL, map_info.size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, map_info.index * getpagesize());
	if (!map) {
		error_msg("UIO map failed for %s (%s)",
		       dev_name, strerror(errno));
		goto close_n_exit;
	}
	regp = map + map_info.offset;

	switch (op) {
	case assign_op:
		*(uint32_t *)regp = val;;
		break;
	case and_op:
		*(uint32_t *)regp &= val;
		break;
	case or_op:
		*(uint32_t *)regp |= val;
		break;
	default:
		ret_val = -1;
		break;
	}

	ret_val = 0;
close_n_exit:
	if (map > 0) munmap (map, map_info.size);

	return ret_val;
}
int uioutil_read_reg (char *class_name, char *uio_name, int fd, uint32_t *val)
{
	char dev_name[MAX_FILE_NAME_LENGTH];
	uioutil_map_info_t map_info;
	int ret_val = -1;
	char *map = 0;
	char *regp;

	if (uioutil_mapinfo_from_name (&map_info, class_name, uio_name) < 0) {

		error_msg("Failed to get mapinfo for name %s class %s",
		       uio_name, class_name);
		goto close_n_exit;
	}
	map = mmap(NULL, map_info.size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, map_info.index * getpagesize());
	if (!map) {
		error_msg("UIO map failed for %s (%s)",
		       dev_name, strerror(errno));
		goto close_n_exit;
	}

	regp = map + map_info.offset;

	*val = *(uint32_t *)regp;

	ret_val = 0;
close_n_exit:
	if (map > 0) munmap (map, map_info.size);

	return ret_val;
}

#endif
